/**
 *  Sankey Diagram Generator
 *  by Sam Brenner
 *  February 23, 2009
 *  v0.6
 */


class BranchGroup
{
  float startY;
  proxml.XMLElement baseElement;
  
  float[] baseValues;
  float[] startHalfPoints;
  float[] endHalfPoints;
  String[] baseText;
  
  float lowValue, highValue;
  int numChildren, baseLevel, elementID;
  color strokeColor;
  String precursorString, passonString;
  
  BranchGroup( proxml.XMLElement element, int level, float sY, int elID, String pString )
  {
    //OFFSETS: To avoid confusion, only add the offset at the very end, inside or directly before the call to bezier().
    startY = sY;            // y axis offset. add to all y values when drawing bezier
    baseLevel = level;      // multiplier for x axis. multiply by branchWidth and add to all x values when drawing bezier.
    deepestLevel = ( baseLevel > deepestLevel ) ? baseLevel : deepestLevel;
    precursorString = pString;
    
    baseElement = element;  // the element containing the node to be drawn and POTENTIALLY any nested nodes.
    elementID = elID;       // the level-zero node which this subnode may belong to. for example, the first <arm> node in the xml will have a elID of 0, and so on.
    numChildren = baseElement.countChildren();
    lowValue = 99999.0;
    highValue = 0.0;
    
    //println( baseElement.getAttribute( "name" ) );
    
    gatherValues();
    generateHalfPoints();
    drawSankey();
    gatherChildren();
    checkForEnd();
  }
  
  void gatherValues()
  {
    //pull only the child node data and place into an array to be drawn.
    
    baseValues = new float[ numChildren ];
    baseText = new String[ numChildren ];
    
    for( int i = 0; i < numChildren; i++ )
    {
       baseValues[ i ] = baseElement.getChild( i ).getFloatAttribute( "value" ) * valueMultiplier;
       baseText[ i ] = baseElement.getChild( i ).getAttribute( "name" );
    } 
  }
  
  void generateHalfPoints()
  {
    //because strokeWeight grows off of the center of the line, the start and end y values of each curve is actually half of its value.
    float runningHeight = 0.0;
    
    startHalfPoints = new float[ numChildren ];
    endHalfPoints = new float[ numChildren ];
     
    for( int i = 0; i < numChildren; i++ )
    {
      startHalfPoints[ i ] = runningHeight + ( baseValues[ i ] / 2 );
      runningHeight += baseValues[ i ];
      
      //when we call map() to calculate the end points of the curves, we will need to know the highest and lowest halfpoint value.      
      lowValue = ( startHalfPoints[ i ] < lowValue ) ? startHalfPoints[ i ] : lowValue;
      highValue = ( startHalfPoints[ i ] > highValue ) ? startHalfPoints[ i ] : highValue;
    }
    
    for( int j = 0; j < numChildren; j++ )
    {
      //curveOffset should be specified in setup(). it causes the outward growth along the y axis.
      if( baseLevel == 0 )
      {
        endHalfPoints[ j ] = map( startHalfPoints[ j ], lowValue, highValue, lowValue - curveOffset, highValue + curveOffset );
      }
      else
      {
        endHalfPoints[ j ] = map( startHalfPoints[ j ], lowValue, highValue, lowValue - altCurveOffset, highValue + altCurveOffset );
      }
    }
  }
  
  void drawSankey()
  {
    //drawing code
    for( int i = 0; i < numChildren; i ++ )
    {
      if( baseLevel == 0 ) 
      {
        stroke( colorArray[i], 80 );
      }
      else
      {
        stroke( colorArray[elementID], 80 );
      }
      
      strokeWeight( baseValues[ i ] );
      strokeCap( SQUARE );
      noFill();
      
      float x1, y1, cx1, cy1, cx2, cy2, x2, y2;
      x1 = startX + ( baseLevel * branchWidth );
      y1 = startHalfPoints[ i ] + startY;
      cx1 = x1 + bezierOffset;
      cy1 = y1;
      
      x2 = x1 + branchWidth;
      y2 = endHalfPoints[ i ] + startY;
      cx2 = x2 - bezierOffset;
      cy2 = y2;
      
      if( numChildren != 1 )
      {
        bezier( x1-1, y1, cx1, cy1, cx2, cy2, x2, y2 );
        
        if( baseElement.getChild( i ).countChildren() == 0 )
        {
          endText.add( precursorString + " " + baseText[ i ] + ": $" + int( ( baseValues[ i ] / valueMultiplier ) / 1000 ) );
          endStartPoints.add( new PVector( x2, y2 ) );
        }
        else
        {
          if( baseElement.getAttribute("name") != "ignore" ) 
          {
            passonString = ( baseElement.getAttribute("name") != precursorString ) ? precursorString + ": " + baseElement.getAttribute( "name" ) : baseElement.getAttribute( "name" );
          }
        }
      }
      else
      {
        //line( x1, y1, x2, y1 );
      }
    }
       
  }
  
  void gatherChildren()
  {
    //go through baseElement again. this time, find which child nodes have children of their own and create a new BranchGroup with that
    //element. As the startY value, pass the element's corresponding endHalfPoints[] value MINUS half of its corresponding baseValues[]
    
    for( int i = 0; i < numChildren; i ++ )
    {
      if( baseElement.getChild( i ).countChildren() != 0 )
      {
        float sY = endHalfPoints[ i ] - ( baseValues[ i ] / 2 );
        int nextElID = ( baseLevel == 0 ) ? i : elementID;
        
        String nextPassonString = ( baseLevel == 0 ) ? baseElement.getChild( i ).getAttribute( "name" ) : passonString;
                
        BranchGroup temp = new BranchGroup( baseElement.getChild( i ), baseLevel + 1, startY + sY, nextElID, nextPassonString );
      }
    }
  }
  
  void checkForEnd()
  {
    childCounter++;
    
    println( childCounter );
    if( childCounter == 76 )
    {
      println( "done" );
      endTextDrawer = new EndTextDrawer();
    }
  }  
}
